@@ -1,9 +1,10 @@
 import asyncio
 import hashlib
 import logging
+from contextlib import AsyncExitStack, asynccontextmanager
+from os import listdir
 from os.path import join
-from posix import listdir
-from typing import Dict, List, cast
+from typing import Dict, List, Optional, cast
 
 from genericpath import exists
 from tortoise import Tortoise
@@ -11,12 +12,12 @@ from tortoise.exceptions import OperationalError
 from tortoise.transactions import get_connection
 from tortoise.utils import get_schema_sql
 
-import dipdup.utils as utils
 from dipdup.codegen import DipDupCodeGenerator
 from dipdup.config import (
     ROLLBACK_HANDLER,
     BcdDatasourceConfig,
     BigMapIndexConfig,
+    CoinbaseDatasourceConfig,
     DatasourceConfigT,
     DipDupConfig,
     IndexConfigTemplateT,
@@ -28,20 +29,26 @@ from dipdup.config import (
 from dipdup.context import DipDupContext, RollbackHandlerContext
 from dipdup.datasources import DatasourceT
 from dipdup.datasources.bcd.datasource import BcdDatasource
+from dipdup.datasources.coinbase.datasource import CoinbaseDatasource
 from dipdup.datasources.datasource import IndexDatasource
 from dipdup.datasources.tzkt.datasource import TzktDatasource
-from dipdup.exceptions import ConfigurationError, HandlerImportError
-from dipdup.hasura import configure_hasura
+from dipdup.exceptions import ConfigurationError, ReindexingRequiredError
+from dipdup.hasura import HasuraGateway
 from dipdup.index import BigMapIndex, Index, OperationIndex
-from dipdup.models import BigMapData, IndexType, OperationData, State
+from dipdup.models import BigMapData, HeadBlockData, IndexType, OperationData, State
+from dipdup.utils import FormattedLogger, slowdown, tortoise_wrapper
+
+INDEX_DISPATCHER_INTERVAL = 1.0
+from dipdup.scheduler import add_job, create_scheduler
 
 
 class IndexDispatcher:
     def __init__(self, ctx: DipDupContext) -> None:
         self._ctx = ctx
 
-        self._logger = logging.getLogger(__name__)
+        self._logger = logging.getLogger('dipdup')
         self._indexes: Dict[str, Index] = {}
+        self._stopped: bool = False
 
     async def add_index(self, index_config: IndexConfigTemplateT) -> None:
         if index_config.name in self._indexes:
@@ -82,12 +89,12 @@ class IndexDispatcher:
 
         self._ctx.reset()
 
-    async def dispatch_operations(self, datasource: TzktDatasource, operations: List[OperationData]) -> None:
+    async def dispatch_operations(self, datasource: TzktDatasource, operations: List[OperationData], block: HeadBlockData) -> None:
         assert len(set(op.level for op in operations)) == 1
         level = operations[0].level
         for index in self._indexes.values():
             if isinstance(index, OperationIndex) and index.datasource == datasource:
-                index.push(level, operations)
+                index.push(level, operations, block)
 
     async def dispatch_big_maps(self, datasource: TzktDatasource, big_maps: List[BigMapData]) -> None:
         assert len(set(op.level for op in big_maps)) == 1
@@ -97,7 +104,19 @@ class IndexDispatcher:
                 index.push(level, big_maps)
 
     async def _rollback(self, datasource: TzktDatasource, from_level: int, to_level: int) -> None:
-        logger = utils.FormattedLogger(ROLLBACK_HANDLER)
+        logger = FormattedLogger(ROLLBACK_HANDLER)
+        if from_level - to_level == 1:
+            # NOTE: Single level rollbacks are processed at Index level.
+            # NOTE: Notify all indexes with rolled back datasource to skip next level and just verify it
+            for index in self._indexes.values():
+                if index.datasource == datasource:
+                    # NOTE: Continue to rollback with handler
+                    if not isinstance(index, OperationIndex):
+                        break
+                    await index.single_level_rollback(from_level)
+            else:
+                return
+
         rollback_fn = self._ctx.config.get_rollback_fn()
         ctx = RollbackHandlerContext(
             config=self._ctx.config,
@@ -120,18 +139,19 @@ class IndexDispatcher:
 
         self._ctx.commit()
 
-        while True:
+        while not self._stopped:
             await self.reload_config()
 
-            # FIXME: Process all indexes in parallel, blocked by https://github.com/tortoise/tortoise-orm/issues/792
-            async with utils.slowdown(1):
-                for index in self._indexes.values():
-                    await index.process()
+            async with slowdown(INDEX_DISPATCHER_INTERVAL):
+                await asyncio.gather(*[index.process() for index in self._indexes.values()])
 
             # TODO: Continue if new indexes are spawned from origination
             if oneshot:
                 break
 
+    def stop(self) -> None:
+        self._stopped = True
+
 
 class DipDup:
     """Main indexer class.
@@ -139,7 +159,7 @@ class DipDup:
     Spawns datasources, registers indexes, passes handler callbacks to executor"""
 
     def __init__(self, config: DipDupConfig) -> None:
-        self._logger = logging.getLogger(__name__)
+        self._logger = logging.getLogger('dipdup')
         self._config = config
         self._datasources: Dict[str, DatasourceT] = {}
         self._datasources_by_config: Dict[DatasourceConfigT, DatasourceT] = {}
@@ -148,21 +168,21 @@ class DipDup:
             datasources=self._datasources,
         )
         self._index_dispatcher = IndexDispatcher(self._ctx)
+        self._scheduler = create_scheduler()
+        self._codegen = DipDupCodeGenerator(self._config, self._datasources_by_config)
 
     async def init(self) -> None:
         """Create new or update existing dipdup project"""
         await self._create_datasources()
 
-        codegen = DipDupCodeGenerator(self._config, self._datasources_by_config)
-        await codegen.create_package()
-        await codegen.fetch_schemas()
-        await codegen.generate_types()
-        await codegen.generate_default_handlers()
-        await codegen.generate_user_handlers()
-        await codegen.cleanup()
+        async with AsyncExitStack() as stack:
+            for datasource in self._datasources.values():
+                await stack.enter_async_context(datasource)
 
-        for datasource in self._datasources.values():
-            await datasource.close_session()
+            await self._codegen.init()
+
+    async def docker_init(self, image: str, tag: str, env_file: str) -> None:
+        await self._codegen.docker_init(image, tag, env_file)
 
     async def run(self, reindex: bool, oneshot: bool) -> None:
         """Main entrypoint"""
@@ -170,17 +190,35 @@ class DipDup:
         url = self._config.database.connection_string
         models = f'{self._config.package}.models'
 
-        async with utils.tortoise_wrapper(url, models):
+        await self._create_datasources()
+
+        hasura_gateway: Optional[HasuraGateway]
+        if self._config.hasura:
+            if not isinstance(self._config.database, PostgresDatabaseConfig):
+                raise RuntimeError
+            hasura_gateway = HasuraGateway(self._config.package, self._config.hasura, self._config.database)
+        else:
+            hasura_gateway = None
+
+        async with AsyncExitStack() as stack:
+            worker_tasks = []
+            await stack.enter_async_context(tortoise_wrapper(url, models))
+            for datasource in self._datasources.values():
+                await stack.enter_async_context(datasource)
+            if hasura_gateway:
+                await stack.enter_async_context(hasura_gateway)
+                worker_tasks.append(asyncio.create_task(hasura_gateway.configure()))
+
             await self._initialize_database(reindex)
-            await self._create_datasources()
             await self._configure()
 
             self._logger.info('Starting datasources')
             datasource_tasks = [] if oneshot else [asyncio.create_task(d.run()) for d in self._datasources.values()]
-            worker_tasks = []
 
-            if self._config.hasura:
-                worker_tasks.append(asyncio.create_task(configure_hasura(self._config)))
+            if self._config.jobs and not oneshot:
+                await stack.enter_async_context(self._scheduler_context())
+                for job_name, job_config in self._config.jobs.items():
+                    add_job(self._ctx, self._scheduler, job_name, job_config)
 
             worker_tasks.append(asyncio.create_task(self._index_dispatcher.run(oneshot)))
 
@@ -188,26 +226,23 @@ class DipDup:
                 await asyncio.gather(*datasource_tasks, *worker_tasks)
             except KeyboardInterrupt:
                 pass
-            finally:
-                self._logger.info('Closing datasource sessions')
-                await asyncio.gather(*[d.close_session() for d in self._datasources.values()])
+            except GeneratorExit:
+                pass
 
-    async def migrate(self) -> None:
+    async def migrate_to_v10(self) -> None:
         codegen = DipDupCodeGenerator(self._config, self._datasources_by_config)
         await codegen.generate_default_handlers(recreate=True)
-        await codegen.migrate_user_handlers_to_v1()
-        self._logger.warning('==================== WARNING =====================')
-        self._logger.warning('Your handlers have just been migrated to v1.0.0 format.')
-        self._logger.warning('Review and commit changes before proceeding.')
-        self._logger.warning('==================== WARNING =====================')
-        quit()
+        await codegen.migrate_user_handlers_to_v10()
+        self._finish_migration('1.0')
+
+    async def migrate_to_v11(self) -> None:
+        codegen = DipDupCodeGenerator(self._config, self._datasources_by_config)
+        await codegen.migrate_user_handlers_to_v11()
+        self._finish_migration('1.1')
 
     async def _configure(self) -> None:
         """Run user-defined initial configuration handler"""
-        try:
-            configure_fn = self._config.get_configure_fn()
-        except HandlerImportError:
-            await self.migrate()
+        configure_fn = self._config.get_configure_fn()
         await configure_fn(self._ctx)
         self._config.initialize()
 
@@ -220,22 +255,25 @@ class DipDup:
             if isinstance(datasource_config, TzktDatasourceConfig):
                 datasource = TzktDatasource(
                     url=datasource_config.url,
-                    cache=self._config.cache_enabled,
+                    http_config=datasource_config.http,
                 )
-                self._datasources[name] = datasource
-                self._datasources_by_config[datasource_config] = datasource
-
             elif isinstance(datasource_config, BcdDatasourceConfig):
                 datasource = BcdDatasource(
-                    datasource_config.url,
-                    datasource_config.network,
-                    self._config.cache_enabled,
+                    url=datasource_config.url,
+                    network=datasource_config.network,
+                    http_config=datasource_config.http,
+                )
+            elif isinstance(datasource_config, CoinbaseDatasourceConfig):
+                datasource = CoinbaseDatasource(
+                    http_config=datasource_config.http,
                 )
-                self._datasources[name] = datasource
-                self._datasources_by_config[datasource_config] = datasource
             else:
                 raise NotImplementedError
 
+            datasource.set_user_agent(self._config.package)
+            self._datasources[name] = datasource
+            self._datasources_by_config[datasource_config] = datasource
+
     async def _initialize_database(self, reindex: bool = False) -> None:
         self._logger.info('Initializing database')
 
@@ -243,30 +281,39 @@ class DipDup:
             await Tortoise._connections['default'].execute_script(f"CREATE SCHEMA IF NOT EXISTS {self._config.database.schema_name}")
             await Tortoise._connections['default'].execute_script(f"SET search_path TO {self._config.database.schema_name}")
 
+        if reindex:
+            self._logger.warning('Started with `--reindex` argument, reindexing')
+            await self._ctx.reindex()
+
+        self._logger.info('Checking database schema')
         connection_name, connection = next(iter(Tortoise._connections.items()))
+
+        try:
+            schema_state = await State.get_or_none(index_type=IndexType.schema, index_name=connection_name)
+        except OperationalError:
+            schema_state = None
+        # TODO: Process exception in Tortoise
+        except KeyError as e:
+            raise ReindexingRequiredError(None) from e
+
         schema_sql = get_schema_sql(connection, False)
 
         # NOTE: Column order could differ in two generated schemas for the same models, drop commas and sort strings to eliminate this
         processed_schema_sql = '\n'.join(sorted(schema_sql.replace(',', '').split('\n'))).encode()
         schema_hash = hashlib.sha256(processed_schema_sql).hexdigest()
 
-        # TODO: Move higher
-        if reindex:
-            self._logger.warning('Started with `--reindex` argument, reindexing')
-            await self._ctx.reindex()
-
-        try:
-            schema_state = await State.get_or_none(index_type=IndexType.schema, index_name=connection_name)
-        except OperationalError:
-            schema_state = None
-
+        # NOTE: `State.index_hash` field contains schema hash when `index_type` is `IndexType.schema`
         if schema_state is None:
             await Tortoise.generate_schemas()
             await self._execute_sql_scripts(reindex=True)
 
-            schema_state = State(index_type=IndexType.schema, index_name=connection_name, hash=schema_hash)
+            schema_state = State(
+                index_type=IndexType.schema,
+                index_name=connection_name,
+                index_hash=schema_hash,
+            )
             await schema_state.save()
-        elif schema_state.hash != schema_hash:
+        elif schema_state.index_hash != schema_hash:
             self._logger.warning('Schema hash mismatch, reindexing')
             await self._ctx.reindex()
 
@@ -298,3 +345,17 @@ class DipDup:
 
             self._logger.info('Executing `%s`', filename)
             await get_connection(None).execute_script(sql)
+
+    def _finish_migration(self, version: str) -> None:
+        self._logger.warning('==================== WARNING =====================')
+        self._logger.warning('Your project has been migrated to spec version %s.', version)
+        self._logger.warning('Review and commit changes before proceeding.')
+        self._logger.warning('==================== WARNING =====================')
+
+    @asynccontextmanager
+    async def _scheduler_context(self):
+        self._scheduler.start()
+        try:
+            yield
+        finally:
+            self._scheduler.shutdown()
