@@ -1,9 +1,11 @@
 import logging
 from abc import abstractmethod
-from collections import deque, namedtuple
+from collections import defaultdict, deque, namedtuple
 from contextlib import suppress
 from typing import Deque, Dict, List, Optional, Set, Tuple, Union, cast
 
+from pydantic.error_wrappers import ValidationError
+
 from dipdup.config import (
     BigMapHandlerConfig,
     BigMapIndexConfig,
@@ -18,19 +20,23 @@ from dipdup.config import (
 )
 from dipdup.context import DipDupContext, HandlerContext
 from dipdup.datasources.tzkt.datasource import BigMapFetcher, OperationFetcher, TzktDatasource
-from dipdup.models import BigMapAction, BigMapData, BigMapDiff, OperationData, Origination, State, TemporaryState, Transaction
+from dipdup.exceptions import InvalidDataError
+from dipdup.models import BigMapData, BigMapDiff, HeadBlockData, OperationData, Origination, State, TemporaryState, Transaction
 from dipdup.utils import FormattedLogger, in_global_transaction
 
-OperationGroup = namedtuple('OperationGroup', ('hash', 'counter'))
+# NOTE: Operations of a single contract call
+OperationSubgroup = namedtuple('OperationSubgroup', ('hash', 'counter'))
 
 
 class Index:
+    _queue: Deque
+
     def __init__(self, ctx: DipDupContext, config: IndexConfigTemplateT, datasource: TzktDatasource) -> None:
         self._ctx = ctx
         self._config = config
         self._datasource = datasource
 
-        self._logger = logging.getLogger(__name__)
+        self._logger = logging.getLogger('dipdup.index')
         self._state: Optional[State] = None
 
     @property
@@ -47,20 +53,21 @@ class Index:
         state = await self.get_state()
         if self._config.last_block:
             last_level = self._config.last_block
-            await self._synchronize(last_level)
+            await self._synchronize(last_level, cache=True)
         elif self._datasource.sync_level is None:
             self._logger.info('Datasource is not active, sync to the latest block')
-            last_level = (await self._datasource.get_latest_block())['level']
+            last_level = (await self._datasource.get_head_block()).level
             await self._synchronize(last_level)
         elif self._datasource.sync_level > state.level:
             self._logger.info('Index is behind datasource, sync to datasource level')
+            self._queue.clear()
             last_level = self._datasource.sync_level
             await self._synchronize(last_level)
         else:
             await self._process_queue()
 
     @abstractmethod
-    async def _synchronize(self, last_level: int) -> None:
+    async def _synchronize(self, last_level: int, cache: bool = False) -> None:
         ...
 
     @abstractmethod
@@ -69,7 +76,7 @@ class Index:
 
     async def _initialize_index_state(self) -> None:
         self._logger.info('Getting state for index `%s`', self._config.name)
-        index_hash = self._config.hash()
+        index_config_hash = self._config.hash()
         state = await State.get_or_none(
             index_name=self._config.name,
             index_type=self._config.kind,
@@ -79,15 +86,25 @@ class Index:
             state = state_cls(
                 index_name=self._config.name,
                 index_type=self._config.kind,
-                hash=index_hash,
+                index_hash=index_config_hash,
                 level=self._config.first_block,
             )
-            await state.save()
 
-        elif state.hash != index_hash:
-            self._logger.warning('Config hash mismatch, reindexing')
+        elif state.index_hash != index_config_hash:
+            self._logger.warning('Config hash mismatch (config has been changed), reindexing')
             await self._ctx.reindex()
 
+        # NOTE: No need to check genesis block
+        if state.level:
+            block = await self._datasource.get_block(state.level)
+            if state.hash:
+                if state.hash != block.hash:
+                    self._logger.warning('Block hash mismatch (missed rollback while dipdup was stopped), reindexing')
+                    await self._ctx.reindex()
+            else:
+                state.hash = block.hash  # type: ignore
+
+        await state.save()
         self._state = state
 
 
@@ -96,11 +113,17 @@ class OperationIndex(Index):
 
     def __init__(self, ctx: DipDupContext, config: OperationIndexConfig, datasource: TzktDatasource) -> None:
         super().__init__(ctx, config, datasource)
-        self._queue: Deque[Tuple[int, List[OperationData]]] = deque()
+        self._queue: Deque[Tuple[int, List[OperationData], Optional[HeadBlockData]]] = deque()
         self._contract_hashes: Dict[str, Tuple[str, str]] = {}
+        self._rollback_level: Optional[int] = None
+        self._last_hashes: Set[str] = set()
 
-    def push(self, level: int, operations: List[OperationData]) -> None:
-        self._queue.append((level, operations))
+    def push(self, level: int, operations: List[OperationData], block: Optional[HeadBlockData] = None) -> None:
+        self._queue.append((level, operations, block))
+
+    async def single_level_rollback(self, from_level: int) -> None:
+        """Ensure next arrived block is the same as rolled back one"""
+        self._rollback_level = from_level
 
     async def _process_queue(self) -> None:
         if not self._queue:
@@ -108,10 +131,10 @@ class OperationIndex(Index):
         self._logger.info('Processing websocket queue')
         with suppress(IndexError):
             while True:
-                level, operations = self._queue.popleft()
-                await self._process_level_operations(level, operations)
+                level, operations, block = self._queue.popleft()
+                await self._process_level_operations(level, operations, block)
 
-    async def _synchronize(self, last_level: int) -> None:
+    async def _synchronize(self, last_level: int, cache: bool = False) -> None:
         """Fetch operations via Fetcher and pass to message callback"""
         state = await self.get_state()
         first_level = state.level
@@ -131,6 +154,7 @@ class OperationIndex(Index):
             last_level=last_level,
             transaction_addresses=transaction_addresses,
             origination_addresses=origination_addresses,
+            cache=cache,
         )
 
         async for level, operations in fetcher.fetch_operations_by_level():
@@ -139,16 +163,39 @@ class OperationIndex(Index):
         state.level = last_level  # type: ignore
         await state.save()
 
-    async def _process_level_operations(self, level: int, operations: List[OperationData]):
+    async def _process_level_operations(self, level: int, operations: List[OperationData], block: Optional[HeadBlockData] = None) -> None:
         state = await self.get_state()
-        if state.level >= level:
-            raise RuntimeError(state.level, level)
+        if level < state.level:
+            raise RuntimeError(f'Level of operation batch is lower than index state level: {level} < {state.level}')
+
+        if self._rollback_level:
+            if state.level != self._rollback_level:
+                raise RuntimeError(f'Rolling back to level {self._rollback_level}, state level {state.level}')
+            if level != self._rollback_level:
+                raise RuntimeError(f'Rolling back to level {self._rollback_level}, got operations of level {level}')
+
+            self._logger.info('Rolling back to previous level, verifying processed operations')
+            expected_hashes = set(self._last_hashes)
+            received_hashes = set([op.hash for op in operations])
+            reused_hashes = received_hashes & expected_hashes
+            if reused_hashes != expected_hashes:
+                self._logger.warning('Attempted a single level rollback but arrived block differs from processed one')
+                await self._ctx.reindex()
+
+            self._rollback_level = None
+            self._last_hashes = set()
+            new_hashes = received_hashes - expected_hashes
+            if not new_hashes:
+                return
+            operations = [op for op in operations if op.hash in new_hashes]
 
         async with in_global_transaction():
             self._logger.info('Processing %s operations of level %s', len(operations), level)
             await self._process_operations(operations)
 
             state.level = level  # type: ignore
+            if block:
+                state.hash = block.hash  # type: ignore
             await state.save()
 
     async def _match_operation(self, pattern_config: OperationHandlerPatternConfigT, operation: OperationData) -> bool:
@@ -185,15 +232,15 @@ class OperationIndex(Index):
             raise NotImplementedError
 
     async def _process_operations(self, operations: List[OperationData]) -> None:
-        """Try to match operations in cache with all patterns from indexes."""
-        operation_groups: Dict[OperationGroup, List[OperationData]] = {}
+        """Try to match operations in cache with all patterns from indexes. Must be wrapped in transaction."""
+        self._last_hashes = set()
+        operation_subgroups: Dict[OperationSubgroup, List[OperationData]] = defaultdict(list)
         for operation in operations:
-            key = OperationGroup(operation.hash, operation.counter)
-            if key not in operation_groups:
-                operation_groups[key] = []
-            operation_groups[key].append(operation)
+            key = OperationSubgroup(operation.hash, operation.counter)
+            operation_subgroups[key].append(operation)
+            self._last_hashes.add(operation.hash)
 
-        for operation_group, operations in operation_groups.items():
+        for operation_subgroup, operations in operation_subgroups.items():
             self._logger.debug('Matching %s', key)
 
             for handler_config in self._config.handlers:
@@ -225,19 +272,19 @@ class OperationIndex(Index):
                         operation_idx += 1
 
                     if pattern_idx == len(handler_config.pattern):
-                        self._logger.info('%s: `%s` handler matched!', operation_group.hash, handler_config.callback)
-                        await self._on_match(operation_group, handler_config, matched_operations)
+                        self._logger.info('%s: `%s` handler matched!', operation_subgroup.hash, handler_config.callback)
+                        await self._on_match(operation_subgroup, handler_config, matched_operations)
 
                         matched_operations = []
                         pattern_idx = 0
 
                 if len(matched_operations) >= sum(map(lambda x: 0 if x.optional else 1, handler_config.pattern)):
-                    self._logger.info('%s: `%s` handler matched!', operation_group.hash, handler_config.callback)
-                    await self._on_match(operation_group, handler_config, matched_operations)
+                    self._logger.info('%s: `%s` handler matched!', operation_subgroup.hash, handler_config.callback)
+                    await self._on_match(operation_subgroup, handler_config, matched_operations)
 
     async def _on_match(
         self,
-        operation_group: OperationGroup,
+        operation_subgroup: OperationSubgroup,
         handler_config: OperationHandlerConfig,
         matched_operations: List[Optional[OperationData]],
     ):
@@ -253,7 +300,15 @@ class OperationIndex(Index):
                     continue
 
                 parameter_type = pattern_config.parameter_type_cls
-                parameter = parameter_type.parse_obj(operation.parameter_json) if parameter_type else None
+                try:
+                    parameter = parameter_type.parse_obj(operation.parameter_json) if parameter_type else None
+                except ValidationError as e:
+                    error_context = dict(
+                        hash=operation.hash,
+                        counter=operation.counter,
+                        nonce=operation.nonce,
+                    )
+                    raise InvalidDataError(operation.parameter_json, parameter_type, error_context) from e
 
                 storage_type = pattern_config.storage_type_cls
                 storage = operation.get_merged_storage(storage_type)
@@ -280,7 +335,7 @@ class OperationIndex(Index):
 
         logger = FormattedLogger(
             name=handler_config.callback,
-            fmt=operation_group.hash + ': {}',
+            fmt=operation_subgroup.hash + ': {}',
         )
         handler_context = HandlerContext(
             datasources=self._ctx.datasources,
@@ -318,7 +373,7 @@ class OperationIndex(Index):
                             strict=pattern_config.strict,
                         ):
                             addresses.add(address)
-        return set(addresses)
+        return addresses
 
     async def _get_contract_hashes(self, address: str) -> Tuple[str, str]:
         if address not in self._contract_hashes:
@@ -346,7 +401,7 @@ class BigMapIndex(Index):
                 level, big_maps = self._queue.popleft()
                 await self._process_level_big_maps(level, big_maps)
 
-    async def _synchronize(self, last_level: int) -> None:
+    async def _synchronize(self, last_level: int, cache: bool = False) -> None:
         """Fetch operations via Fetcher and pass to message callback"""
         state = await self.get_state()
         first_level = state.level
@@ -366,6 +421,7 @@ class BigMapIndex(Index):
             last_level=last_level,
             big_map_addresses=big_map_addresses,
             big_map_paths=big_map_paths,
+            cache=cache,
         )
 
         async for level, big_maps in fetcher.fetch_big_maps_by_level():
@@ -400,17 +456,24 @@ class BigMapIndex(Index):
         matched_big_map: BigMapData,
     ) -> None:
         """Prepare handler arguments, parse key and value. Schedule callback in executor."""
-        if matched_big_map.action == BigMapAction.ALLOCATE:
-            return
 
-        key_type = handler_config.key_type_cls
-        key = key_type.parse_obj(matched_big_map.key)
-
-        if matched_big_map.action == BigMapAction.REMOVE:
-            value = None
+        if matched_big_map.action.has_key:
+            key_type = handler_config.key_type_cls
+            try:
+                key = key_type.parse_obj(matched_big_map.key)
+            except ValidationError as e:
+                raise InvalidDataError(matched_big_map.key, key_type) from e
         else:
+            key = None
+
+        if matched_big_map.action.has_value:
             value_type = handler_config.value_type_cls
-            value = value_type.parse_obj(matched_big_map.value)
+            try:
+                value = value_type.parse_obj(matched_big_map.value)
+            except ValidationError as e:
+                raise InvalidDataError(matched_big_map.key, value_type) from e
+        else:
+            value = None
 
         big_map_context = BigMapDiff(  # type: ignore
             data=matched_big_map,
